# QuantumLattices.jl

[![CI][main-ci-img]][main-ci-url]
[![codecov][codecov-img]][codecov-url]
[![][docs-latest-img]][docs-latest-url]
[![][docs-stable-img]][docs-stable-url]
[![996.icu](https://img.shields.io/badge/link-996.icu-red.svg)](https://996.icu)
[![LICENSE](https://img.shields.io/badge/License-Apache%202.0-blue.svg)](https://opensource.org/licenses/Apache-2.0)
[![LICENSE](https://img.shields.io/badge/license-Anti%20996-blue.svg)](https://github.com/996icu/996.ICU/blob/master/LICENSE)

*Julia package for the construction of quantum lattice systems.*

Welcome to [QuantumLattices](https://github.com/Quantum-Many-Body/QuantumLattices.jl). Here we provide a general framework to construct the **second-quantized operator-formed Hamiltonian of any quantum lattice system**, with the inputs as simple as its description by the natural language. Combined with [SymPy](https://github.com/JuliaPy/SymPy.jl), this operator-formed Hamiltonian supports **complete symbolic computations**, making it a convenient "frontend" of quantum many-body algorithms, such as **TBA**(tight-bind approximation), **LSWT**(linear spin wave theory), **SCMF**(self-consistent mean field theory), **ED**(exact diagonalization), **CPT/VCA**(cluster perturbation theory and variational cluster approach ), **DMRG**(density matrix renormalization group), etc. Generic interfaces are defined to provide a unified access to these algorithms which would be implemented in separate packages. Only minor modifications need be made when users alter from one algorithm to another.

## Installation

In Julia **v1.6+**, please type `]` in the REPL to use the package mode, then type this command:

```julia
pkg> add QuantumLattices
```

## Introduction

The core of the package is the construction of the **operator representation of a lattice Hamiltonian**. This is based on the following mathematical observations that the operators in a lattice Hamiltonian:
* **act on local Hilbert spaces**, and
* **form an algebra over the complex field**.

The first observation is the starting point of our [**unitcell description framework**](https://quantum-many-body.github.io/QuantumLattices.jl/dev/tutorials/UnitcellDescription/Introduction/) and the second is the mathematical foundation of our **symbolic computing system**.

It is noted that our implementation of the symbolic computation only involves
* the mathematical operations between a scalar and an operator, and
* the mathematical operations between two operators.

The symbolic operations between two scalars are **not** implemented because:
* in condensed matter physics, for many cases, only the numerical values of operators are relevant because the analytical expressions could be too complicated to analyze or they may even not exist;
* our construction process of the operators and their mathematical operations are **compatible with the [SymPy](https://github.com/JuliaPy/SymPy.jl) package**, therefore, a fully symbolic computation can be achieved by a simple combination of both.

The major aim of this package is to provide **unified "frontend" to access all quantum many-body algorithms**. Much of the job can be done by the construction of the operator-formed Hamiltonian, which serves as a common input for different algorithms. Utilities are also provided to deal with the project management, such as result recording, data caching, parameter updating, code logging, dependency managing, etc.

## Package Features

* **Unitcell Description Framework**: by telling the information of the quantum lattice system within a unitcell, the construction of the symbolic representation of the Hamiltonian is just as simple as describing the system in a usual research paper.
* **Complete Symbolic Computation**: with only this package, symbolic computation between operators is realized whereas the coefficient of any operator remains numeric; by integrating it with [SymPy](https://github.com/JuliaPy/SymPy.jl), complete symbolic computation can be achieved and no modifications need be made on the methods in this package.
* **Generic Frontend of Many-Body Algorithms**: quantum many-body algorithms can be initialized in quite similar ways with only minor modifications needed. Moreover, automatic project management is realized, including that of result recording, data caching, parameter updating, code logging, dependency managing, etc.

## Supported Systems

Four common kinds of systems in condensed matter physics are supported:
* **canonical fermionic systems**
* **canonical/hard-core bosonic systems**
* **SU(2) spin systems**
* **Phononic systems**

Furthermore, other systems can be supported easily by extending the generic "protocols" provided in this package.

## Supported Algorithms

Concrete algorithms could be considered as the "backend" of quantum lattice systems. They are developed in separate packages (still in progress):
* **[TBA](https://github.com/Quantum-Many-Body/TightBindingApproximation.jl)**: tight-binding approximation for fermionic/bosonic systems;
* **SCMF**: self-consistent mean field theory for fermionic systems;
* **[ED](https://github.com/Quantum-Many-Body/ExactDiagonalization.jl)**: exact diagonalization for fermionic/hard-core-bosonic/spin systems;
* **CPT/VCA**: cluster perturbation theory and variational cluster approach for fermionic systems;
* **DMRG**: density matrix renormalization group for fermionic/hard-core-bosonic/spin systems;
* **[SWT](https://github.com/Quantum-Many-Body/SpinWaveTheory.jl)**: spin wave theory for local spin systems.

## Getting Started
[Tutorials: unitcell description](https://quantum-many-body.github.io/QuantumLattices.jl/dev/tutorials/UnitcellDescription/Introduction/)

## Documentation
- [**LATEST**][docs-latest-url] &mdash; **documentation of the latest version.**
- [**STABLE**][docs-stable-url] &mdash; **documentation of the stable version.**

## Note

Due to the fast development of this package, releases with different minor version numbers are **not** guaranteed to be compatible with previous ones **before** the release of v1.0.0. Comments are welcomed in the GitHub issues.

## Contact
waltergu1989@gmail.com

## Python counterpart
[HamiltonianPy](https://github.com/waltergu/HamiltonianPy): in fact, the authors of this Julia package worked on the python package at first and only turned to Julia later.

[main-ci-img]: https://github.com/Quantum-Many-Body/QuantumLattices.jl/actions/workflows/CI.yml/badge.svg
[main-ci-url]: https://github.com/Quantum-Many-Body/QuantumLattices.jl/actions/workflows/CI.yml
[codecov-img]: https://codecov.io/gh/Quantum-Many-Body/QuantumLattices.jl/branch/master/graph/badge.svg
[codecov-url]: https://codecov.io/gh/Quantum-Many-Body/QuantumLattices.jl
[docs-latest-img]: https://img.shields.io/badge/docs-latest-blue.svg
[docs-latest-url]: https://quantum-many-body.github.io/QuantumLattices.jl/latest/
[docs-stable-img]: https://img.shields.io/badge/docs-stable-blue.svg
[docs-stable-url]: https://quantum-many-body.github.io/QuantumLattices.jl/stable/
